function drawPtHighlight(ctx, color, coordinate, vradius) {
  ctx.fillStyle = color;
  ctx.strokeStyle = color;
  ctx.lineWidth = VERTEX_LINE_WIDTH;
  ctx.beginPath();
  ctx.arc(coordinate.x, coordinate.y, vradius + HIGHLIGHT_OFFSET, 0, 2 * Math.PI);
  ctx.fill();
  ctx.stroke();
}

function drawPt(ctx, color, coordinate, radius) {
  ctx.fillStyle = color;
  ctx.strokeStyle = VERTEX_LINE_COLOR;
  ctx.lineWidth = VERTEX_LINE_WIDTH;
  ctx.beginPath();
  ctx.arc(coordinate.x, coordinate.y, radius, 0, 2 * Math.PI);
  ctx.fill();
  ctx.stroke();
}
function drawPtLabel(ctx, coordinate, label = "") {
  ctx.fillStyle = VERTEX_LABEL_COLOR;
  ctx.font = PT_FONT;
  ctx.textAlign = "center";
  ctx.textBaseline = 'middle';
  ctx.fillText(label, coordinate.x, coordinate.y);
}

function drawQuadrilateral(ctx, origin, axes, color) {
  ctx.fillStyle = color;
  ctx.beginPath();
  ctx.moveTo(origin.x, origin.y);
  ctx.lineTo(axes[0].x + origin.x, axes[0].y + origin.y);
  ctx.lineTo(axes[0].x + axes[1].x + origin.x, axes[0].y + axes[1].y + origin.y);
  ctx.lineTo(axes[1].x + origin.x, axes[1].y + origin.y);
  ctx.closePath();
  ctx.fill(); 
  
}

function drawPolygon(ctx, vertices, vertexIds, color) {
  ctx.fillStyle = color;
  ctx.beginPath();
  ctx.moveTo(vertices[vertexIds[0]].x, vertices[vertexIds[0]].y);
  for (let i = 1; i < vertexIds.length; i++) {
    ctx.lineTo(vertices[vertexIds[i]].x, vertices[vertexIds[i]].y);
  }
  ctx.closePath();
  ctx.fill();
}

function drawLattice(ctx, lattice) {
  for (let i = 0; i < lattice.length; i++) {
    drawPt(ctx, LATTICE_FILL_COLOR, lattice[i], LATTICE_RADIUS);
  }
}

function drawOrigin(ctx, origin) {
  drawPt(ctx, ORIGIN_FILL_COLOR, origin, ORIGIN_RADIUS);
}

function drawAxes(ctx, origin, axes) {
  for (let i = 0; i < axes.length ; i++){
    //compute endpount of the axis
    var coordinate = new Coordinate(origin.x, origin.y);
    coordinate.add(axes[i]);
    //draw labeled line and head
    drawLabelEdge(ctx, false, origin, coordinate, AXES_LINE_WIDTH, AXES_FILL_COLORS[i], AXES_FILL_COLORS[i], AXES_LABELS[i], true);
    //drawHead(ctx, origin, coordinate);
  }
}
//https://stackoverflow.com/questions/808826/draw-arrow-on-canvas-tag
function drawHead(ctx, coordinate1, coordinate2, color, axesFlag){
  var r = ARROW_HEAD_SIZE;
  var angle, x, y;

  angle = Math.atan2(coordinate2.y - coordinate1.y, coordinate2.x - coordinate1.x);

  var tox = coordinate2.x;
  var toy = coordinate2.y;

  var arrow_offset = VERTEX_RADIUS * .80;
  if(axesFlag) {
    r = ARROW_HEAD_SIZE_SMALL;
    arrow_offset = LATTICE_RADIUS * 2;
  }
  //console.log(arrow_offset);
  tox -= Math.cos(angle) * arrow_offset;
  toy -= Math.sin(angle) * arrow_offset;
  
  ctx.fillStyle = color;
  ctx.beginPath();

  x = r*Math.cos(angle) + tox;
  y = r*Math.sin(angle) + toy;
  
  //angle = Math.atan2(coordinate2.y - coordinate1.y, coordinate2.x - coordinate1.x);
  //x = r*Math.cos(angle) + coordinate2.x;
  //y = r*Math.sin(angle) + coordinate2.y;

  ctx.moveTo(x, y);

  angle += (1/3)*(2*Math.PI);
  x = r*Math.cos(angle) + tox;
  y = r*Math.sin(angle) + toy;

  ctx.lineTo(x, y);

  angle += (1/3)*(2*Math.PI);
  x = r*Math.cos(angle) + tox;
  y = r*Math.sin(angle) + toy;

  ctx.lineTo(x, y);
  ctx.closePath();
  ctx.fill();
}

function drawEdge(ctx, showUndirected, coordinate1, coordinate2, lineWidth, edgeColor, axesFlag) {
  ctx.strokeStyle = edgeColor;
  ctx.lineWidth = lineWidth;
  ctx.beginPath();
  ctx.moveTo(coordinate1.x, coordinate1.y);
  ctx.lineTo(coordinate2.x, coordinate2.y);
  ctx.stroke();
  if (!showUndirected || axesFlag)
    drawHead(ctx, coordinate1, coordinate2, edgeColor, axesFlag);
}

function drawLabelEdge(ctx, showUndirected, coordinate1, coordinate2, lineWidth, edgeColor, labelColor, label = null, axesFlag = false) {
  const distance = Math.round(coordinate1.distance(coordinate2));
  ctx.setLineDash([(distance - WEIGHT_SPACE_LENGTH)/2, WEIGHT_SPACE_LENGTH]);
  drawEdge(ctx, showUndirected, coordinate1, coordinate2, lineWidth, edgeColor, axesFlag);
  ctx.setLineDash([]);

  ctx.fillStyle = labelColor;
  ctx.font = EDGE_FONT;
  ctx.textAlign = "center";
  ctx.textBaseline = 'middle';
  ctx.fillText(label === null ? distance : label, 
    (coordinate1.x + coordinate2.x)/2, 
    (coordinate1.y + coordinate2.y)/2);
}

function drawGraph(canvas, graph, settings) {
  //stateHandler();
  const ctx = canvas.getContext("2d");
  ctx.clearRect(0,0,canvas.width, canvas.height);
  var showUndirected = settings.getDisplay("showUndirected");
  
  if (settings.getDisplay("showUCell") && graph.origin != null && graph.axes.length === 2)
    drawQuadrilateral(ctx, graph.origin, graph.axes, UCELL_COLOR);
  
  if (settings.getDisplay("showLattice") && graph.lattice.length > 0)
    drawLattice(ctx, graph.lattice);
  
  if (settings.getDisplay("showAxes") && graph.axes.length > 0)
    drawAxes(ctx, graph.origin, graph.axes);
 
  //console.log("settings.getDisplay('showOrigin') " + settings.getDisplay("showOrigin"));
  //console.log("graph.origin " + graph.origin);
  if (settings.getDisplay("showOrigin") && graph.origin != null)
    drawOrigin(ctx, graph.origin);
  
  if (settings.getDisplay("showFaces")) {
    for (let i = 0; i < graph.faces.length; i++) {
      var color = FACE_COLORS[graph.forbits[i]];
      if (settings.getDisplay("showFUniform")) color = FACE_DEFAULT_COLOR;
      drawPolygon(ctx, graph.vertices, graph.faces[i], color); 
    }
  }
  
  //if show span
  if (settings.getDisplay("showSpan")) {
    if (settings.getDisplay("showEdges")) {//Draw Edges
      if (settings.getDisplay("showELabels")) {
        var labels = [];
        if (settings.getDisplay("showEIds"))//show edge ids
          labels = [...Array(graph.spanE.length).keys()];
        else if (settings.getDisplay("showEOrbits"))//show edge orbits
          labels = [...Array(graph.spanE.length).keys()];
        else if (settings.getDisplay("showEShifts")) {//show edge shifts
          labels = graph.eRepShifts;
        }
        else if (settings.getDisplay("showELengths"))//show edge lengths
          labels = new Array(graph.spanE.length).fill(null);
        //Draw with labels
        for (let i = 0; i < graph.spanE.length; i++) {
          var color = EDGE_COLORS[i];
          if (settings.getDisplay("showEUniform")) color = EDGE_DEFAULT_COLOR;
          if (i >= EBOXES || document.getElementById("ecolor" + i).checked) 
            drawLabelEdge(ctx, showUndirected, graph.spanV[graph.spanE[i][0]], graph.spanV[graph.spanE[i][1]], EDGE_LINE_WIDTH, color, EDGE_LABEL_COLOR, labels[i]);
        }
      }
      else//Draw without labels
        for (let i = 0; i < graph.spanE.length; i++) {
          var color = EDGE_COLORS[i];
          if (settings.getDisplay("showEUniform")) color = EDGE_DEFAULT_COLOR;
          if (i >= EBOXES || document.getElementById("ecolor" + i).checked) 
            drawEdge(ctx, showUndirected, graph.spanV[graph.spanE[i][0]], graph.spanV[graph.spanE[i][1]], EDGE_LINE_WIDTH, color);
        }
    }
    //Draw Vertices
    if (settings.getDisplay("showVertices")) {
      for (let i = 0; i < graph.spanV.length; i++) {
        var color = VERTEX_COLORS[graph.spanVOrbits[i]];
        if (settings.getDisplay("showVUniform")) color = VERTEX_DEFAULT_COLOR;
        if (graph.spanVOrbits[i] >= VBOXES || document.getElementById("vcolor" + graph.spanVOrbits[i]).checked)
          drawPt(ctx, color, graph.spanV[i], VERTEX_RADIUS);
      }
    }
    //Draw Vertex Labels
    if (settings.getDisplay("showVLabels")) {
      var labels = [];
      if (settings.getDisplay("showVIds"))//show vertex ids
        labels = [...Array(graph.spanV.length).keys()];
      else if (settings.getDisplay("showVOrbits"))//show vertex orbits
        labels = graph.spanVOrbits;
      else if (settings.getDisplay("showVShifts"))//show vertex shifts
        labels = graph.spanVShifts;
      //Draw with labels
      for (let i = 0; i < graph.spanV.length; i++)
        drawPtLabel(ctx, graph.spanV[i],  labels[i]);
    }
  }
  else {
    //Draw Edges
    if (settings.getDisplay("showEdges")) {
      if (settings.getDisplay("showELabels")) {
        //If the labels shown should be IDS
        if (settings.getDisplay("showEIds")) {
          for (let i = 0; i < graph.edges.length; i++)  {
            if (graph.eorbits[i] >= EBOXES || document.getElementById("ecolor" + graph.eorbits[i]).checked) {
              var color = EDGE_COLORS[graph.eorbits[i]];
              if (settings.getDisplay("showEUniform")) color = EDGE_DEFAULT_COLOR;
              drawLabelEdge(ctx, showUndirected, graph.vertices[graph.edges[i][0]], graph.vertices[graph.edges[i][1]], EDGE_LINE_WIDTH, color, EDGE_LABEL_COLOR, i);
            }
          }
        }
        //If the labels shown should be orbits, shifts, or lengths
        else {
          for (let i = 0; i < graph.edges.length; i++) {
            if (graph.eorbits[i] >= EBOXES || document.getElementById("ecolor" + graph.eorbits[i]).checked) {
              var label;
              var color = EDGE_COLORS[graph.eorbits[i]];
              if (settings.getDisplay("showEUniform")) color = EDGE_DEFAULT_COLOR;
              if (isZeroShift(graph.vshifts[graph.edges[i][0]])) {// if it is representative, draw label
                if (settings.getDisplay("showEOrbits"))
                  label = graph.eorbits[i];
                else if (settings.getDisplay("showEShifts"))
                  label = graph.eRepShifts[graph.eorbits[i]];
                else if (settings.getDisplay("showELengths"))
                  label = null;
                drawLabelEdge(ctx, showUndirected, graph.vertices[graph.edges[i][0]], graph.vertices[graph.edges[i][1]], EDGE_LINE_WIDTH, color, EDGE_LABEL_COLOR, label);
              }
              else
                drawEdge(ctx, showUndirected, graph.vertices[graph.edges[i][0]], graph.vertices[graph.edges[i][1]], EDGE_LINE_WIDTH, color);
            }
          }
        }    
      }
      else//Draw without labels
        for (let i = 0; i < graph.edges.length; i++) {
          var color = EDGE_COLORS[graph.eorbits[i]];
          if (settings.getDisplay("showEUniform")) color = EDGE_DEFAULT_COLOR;
          if (graph.eorbits[i] >= EBOXES || document.getElementById("ecolor" + graph.eorbits[i]).checked) 
            drawEdge(ctx, showUndirected, graph.vertices[graph.edges[i][0]], graph.vertices[graph.edges[i][1]], EDGE_LINE_WIDTH, color);
        }
    }
    //Draw Vertices
    if (settings.getDisplay("showVertices")) {
      for (let i = 0; i < graph.vertices.length; i++) {
        var color = VERTEX_COLORS[graph.vorbits[i]];
        if (settings.getDisplay("showVUniform")) color = VERTEX_DEFAULT_COLOR;
        if (graph.vorbits[i] >= VBOXES || document.getElementById("vcolor" + graph.vorbits[i]).checked)
          drawPt(ctx, color, graph.vertices[i], VERTEX_RADIUS);
      }
    }
    //Draw Vertex Labels
    if (settings.getDisplay("showVLabels")) {
      var labels = [];
      if (settings.getDisplay("showVIds"))//show vertex ids
        labels = [...Array(graph.vertices.length).keys()];
      else if (settings.getDisplay("showVOrbits"))//show vertex orbits
        labels = graph.vorbits;
      else if (settings.getDisplay("showVShifts"))//show vertex shifts
        labels = graph.vshifts;
      //Draw labels
      for (let i = 0; i < graph.vertices.length; i++)
        drawPtLabel(ctx, graph.vertices[i],  labels[i]);
    }
  }

  if (~graph.selectedVID && graph.selectedVID != -1) {
    //console.log("graph.selectedVID: " + graph.selectedVID);
    //console.log("settings.getDisplay('showOrigin'): " + settings.getDisplay("showOrigin"));
    if (graph.selectedVID === -2) {
      if (settings.getDisplay("showOrigin"))
        drawOrigin(ctx, graph.origin);
      else
        graph.selectedVID = -1;
    }
    else {
      let currStateVal = document.querySelector('input[name="states"]:checked').value; 
      if (currStateVal === "vertices" || currStateVal === "edges")
        drawPtHighlight(ctx, HIGHLIGHT_COLOR, graph.vertices[graph.selectedVID], VERTEX_RADIUS);
        //drawSelect(ctx, graph.vertices[graph.selectedVID], HIGHLIGHT_COLOR);
    }
  }
  if (~graph.allSelected && graph.allSelected.length != 0) {
      let currStateVal = document.querySelector('input[name="states"]:checked').value; 
      if (currStateVal === "faces") {
        drawPtHighlight(ctx, HIGHLIGHT_COLOR, graph.vertices[graph.allSelected[0]], VERTEX_RADIUS);
        //drawSelect(ctx, graph.vertices[graph.allSelected[0]], HIGHLIGHT_COLOR);
        for (let i = 1; i < graph.allSelected.length; i++)
          drawPtHighlight(ctx, OTHER_HIGHLIGHT_COLOR, graph.vertices[graph.allSelected[i]], VERTEX_RADIUS);
          //drawSelect(ctx, graph.vertices[graph.allSelected[i]], OTHER_HIGHLIGHT_COLOR);
      }
  }
}